/******************************************************************************
* Project:		morsix3330
* File:			disp3330.c
* Date:			20.12.2007
* Author:		Philipp Burch
* E-Mail:		philipp_burch@gmx.ch
* Description:	This file contains Display-IO functions.
******************************************************************************/

#include <disp3330.h>
#include <stdint.h>
#include <util/delay.h>
#include <avr/io.h>
#include <charset.h>
#include <main.h>
#include <string.h>
#include <avr/pgmspace.h>
#include <debug.h>

#if DISP_USE_CACHE == 1
	uint8_t Cache[DISP_WIDTH * DISP_HEIGHT / 8];
#endif


//Private (Helper functions)

#if DISP_USE_CACHE == 1
	inline uint8_t *CacheXY(const uint8_t x, const uint8_t y, uint8_t *mask) {
		#if DISP_CACHE_ALGN == 0
			//Horizontal alignment
			#if DISP_WIDTH % 8 == 0
				*mask = 1 << (x % 8);
			#else
				*mask = 1 << ((x + y * DISP_WIDTH) % 8);
			#endif
			return Cache + (x + y * DISP_WIDTH) / 8;
		#else
			//Vertical alignment
			#if DISP_HEIGHT % 8 == 0
				*mask = 1 << (y % 8);
			#else
				*mask = 1 << ((x * DISP_HEIGHT + y) % 8);
			#endif
			return Cache + ((x * DISP_HEIGHT + y) / 8);
		#endif
	}
#endif

inline void SendSPI(uint8_t data) {
	SPDR = data;
	loop_until_bit_is_set(SPSR, SPIF);
}

inline void UseMask(uint8_t *d, uint8_t mask, draw_mode_t mode) {
	switch (mode) {
		case COPY:
		case OR:
			*d |= mask; break;
		case XOR:
			*d ^= mask; break;
		case CLEAR:
			*d &= ~mask; break;
		default:
			;
	}
}


//Public functions

void Disp_Init() {
	_delay_ms(10);

	RPO_DISP_RES |= _BV(PIN_DISP_RES);

	DISP_SELECT;
	DISP_DC_CMD;
	SendSPI(0x23);	//Startup, Vertical addressing, Extended instruction set
	SendSPI(0x06);	//Temperature control value 2
	SendSPI(0x14);	//Bias system 4
	SendSPI(0xad);	//Vop (Contrast) 50
	SendSPI(0x22);	//Startup, Vertical addressing, Basic instruction set
	SendSPI(0x40);	//Y = 0
	SendSPI(0x80);	//X = 0
	SendSPI(0x08);	//Display blank

	DISP_RELEASE;
	Disp_Clear_Cache();
	Disp_Refresh();
	_delay_ms(10);

	DISP_SELECT;
	DISP_DC_CMD;
	SendSPI(0x22);	//Startup, Vertical addressing, Basic instruction set
	SendSPI(0x0c);	//Normal mode
	DISP_RELEASE;
}

#if DISP_USE_CACHE == 1 || DISP_CAN_READ == 1
void Disp_SetPx(const uint8_t x, const uint8_t y) {
	if (x > DISP_WIDTH - 1 || y > DISP_HEIGHT - 1) return;

	#if DISP_USE_CACHE == 1
		uint8_t mask = 0;
		*CacheXY(x, y, &mask) |= mask;
	#endif
}

void Disp_ClrPx(const uint8_t x, const uint8_t y) {
	if (x > DISP_WIDTH - 1 || y > DISP_HEIGHT - 1) return;

	#if DISP_USE_CACHE == 1
		uint8_t mask = 0;
		*CacheXY(x, y, &mask) &= ~mask;
	#endif
}

void Disp_InvPx(const uint8_t x, const uint8_t y) {
	if (x > DISP_WIDTH - 1 || y > DISP_HEIGHT - 1) return;

	#if DISP_USE_CACHE == 1
		uint8_t mask = 0;
		*CacheXY(x, y, &mask) ^= mask;
	#endif
}

bool Disp_GetPx(const uint8_t x, const uint8_t y) {
	if (x > DISP_WIDTH - 1 || y > DISP_HEIGHT - 1) return false;

	#if DISP_USE_CACHE == 1
		uint8_t mask = 0;
		return *CacheXY(x, y, &mask) & mask;
	#endif
}

point_t Disp_Print(const uint8_t left,
				   const uint8_t top,
				   const uint8_t width,
				   const uint8_t height,
				   const uint8_t x,
				   const uint8_t y,
				   char *s,
				   const draw_mode_t mode,
				   const bool break_word,
				   const bool use_controls) {

	point_t ret = {x, y};
	if (left + width > DISP_WIDTH || top + height > DISP_HEIGHT) return ret;
	if (*s == 0) return ret;
	if (x < left || x > left + width) return ret;
	if (y < top || y > top + height - DISP_LIN_HEIGHT) return ret;

	uint8_t t;
	uint8_t mr = left + width - 1, mb = top + height - 1;
	bool lf = false;
	if (!width) mr = DISP_WIDTH - 1;
	if (!height) mb = DISP_HEIGHT - 1;

	for (;;) {				//Lines
		while (*s) {		//Characters
			if (use_controls && (*s == '\n' || *s == '\r' || *s == '\t')) {
				lf = false;
				//Control characters
				switch (*s) {
					case '\n':
						//Always perform a new line
						lf = true;
						break;

					case '\r':
						//Perform a new line and skip a following line feed
						if (*(s + 1) == '\n') ++s;
						lf = true;
						break;

					case '\t':
						t = (ret.x / DISP_TAB_WIDTH + 1) * DISP_TAB_WIDTH;
						if (t >= mr) {
							lf = true;
						} else {
							ret.x = t;
						}
						break;
				}
				if (lf) {
					++s;
					break;
				}
			} else {
				if (!break_word && *s == ' ') {
					//Check if the next word fits into the line
					uint8_t w = 0;
					lf = false;
					char *c = s + 1;
					while (*c != ' ' && *c) {
						w += (pgm_read_byte(&CS_TOC[(uint8_t)*c++].size_class) & 0x1f) + 1;
						if (!lf && w > mr - ret.x) {
							//It doesn't...
							lf = true;
						} else if (w > width) {
							//It doesn't, but it doesn't even fit into one full line
							lf = false;
							break;
						}
					}
					if (lf) {
						++s;
						break;
					}
				}
				#if DISP_USE_CACHE == 1
					uint16_t o = pgm_read_word(&CS_TOC[(uint8_t)*s].offset);
					uint8_t optimize = 0;
					#if DISP_CACHE_ALGN == 1
						//Vertical addressing
						optimize = 1;
						if (ret.y % 8 == 0) optimize = 2;
					#endif

					uint8_t mask;
					uint8_t w = (pgm_read_byte(&CS_TOC[(uint8_t)*s].size_class) & 0x1f) + 1;

					//Check if the character fits into the line
					if (ret.x + w > mr) break;

					if (optimize == 2) {
						uint8_t *data = CacheXY(ret.x, ret.y, &mask);
						for (uint8_t i = 0; i < w; ++i) {
							switch (mode) {
								case COPY:
									*data = pgm_read_byte(&CS_Data[o]); break;
								case AND:
									*data &= pgm_read_byte(&CS_Data[o]); break;
								case OR:
									*data |= pgm_read_byte(&CS_Data[o]); break;
								case XOR:
									*data ^= pgm_read_byte(&CS_Data[o]); break;
								case CLEAR:
									*data &= ~pgm_read_byte(&CS_Data[o]); break;
							}
							data += DISP_HEIGHT / 8;
							++o;
						}
					} else if (optimize == 1) {
						uint8_t n = ret.y % 8;
						uint8_t *data = CacheXY(ret.x, ret.y, &mask);
						--mask;
						for (uint8_t i = 0; i < w; ++i) {
							uint16_t d = (uint16_t)pgm_read_byte(&CS_Data[o++]) << n;
							switch (mode) {
								case COPY:
									*data = (*data & mask) | L8(d);
									*(data + 1) = (*(data + 1) & ~mask) | H8(d);
									break;
								case AND:
									*data &= L8(d) | mask;
									*(data + 1) &= H8(d) | ~mask;
									break;
								case OR:
									*data |= L8(d);
									*(data + 1) |= H8(d);
									break;
								case XOR:
									*data ^= L8(d);
									*(data + 1) ^= H8(d);
									break;
								case CLEAR:
									*data &= ~L8(d);
									*(data + 1) &= ~H8(d);
									break;
							}
							data += DISP_HEIGHT / 8;
						}
					} else {
						for (uint8_t i = 0; i < w; ++i) {
							uint8_t b = pgm_read_byte(&CS_Data[o++]);
							for (uint8_t j = 0; j < 8; ++j) {
								uint8_t *data = CacheXY(ret.x + i, ret.y + j, &mask);
								switch (mode) {
									case COPY:
										*data = (*data & ~mask) | (b & 1 ? mask : 0); break;
									case AND:
										*data &= (b & 1 ? *data | mask : *data); break;
									case OR:
										*data |= (b & 1 ? mask : 0); break;
									case XOR:
										*data ^= (b & 1 ? mask : 0); break;
									case CLEAR:
										*data = (*data & ~mask) | (b & 1 ? 0 : mask); break;
								}
								b >>= 1;
							}
						}
					}

					ret.x += w + 1;
				#endif //DISP_USE_CACHE == 1
			} //if (use_controls && (*s == '\n' || *s == '\r' || *s == '\t'))
			++s;
		} //while (*s)
		if (*s || lf) {
			ret.y += DISP_LIN_HEIGHT;
			ret.x = left;
		}
		if (!*s) return ret;
		if (ret.y > mb - DISP_LIN_HEIGHT + 1) break;
		lf = false;
	} //for (;;)

	return ret;
}

void Disp_Box(uint8_t x, 
			  uint8_t y, 
			  uint8_t width, 
			  uint8_t height, 
			  draw_mode_t mode, 
			  bool fill) {

	if (x + width > DISP_WIDTH || y + height > DISP_HEIGHT) return;
	if (width < 2 || height < 2) {
		Disp_Line(x, y, x + width - 1, y + height - 1, mode);
		return;
	}

	#if DISP_USE_CACHE == 1
		uint8_t mask1 = 0, mask2 = 0, vmask1, vmask2;
		uint8_t *d1, *d2, *d3;

		d1 = CacheXY(x, y, &mask1);
		d2 = CacheXY(x + width - 1, y, &mask1);
		d3 = CacheXY(x, y + height - 1, &mask2);

		#if DISP_CACHE_ALGN == 1
			//Vertical alignment
			uint8_t n = 0;
			if (d1 != d3) {
				//More than one byte in height
				vmask1 = ~(mask1 - 1) << 1;
				vmask2 = mask2 - 1;
				n = d3 - d1 - 1;
				if (n == 0xff) n = 0;
			} else {
				//Same byte
				vmask1 = ~((mask1 - 1) | mask1 | ~(mask2 - 1));
				vmask2 = 0;
			}

			uint8_t *a, *b;
			uint8_t h = DISP_HEIGHT / 8;
			for (a = d1, b = d3; a <= d2; a += h, b += h) {
				switch (mode) {
					case COPY:
					case OR:
						*a |= mask1 | (fill && a > d1 && a < d2 ? vmask1 : 0);
						*b |= mask2 | (fill && a > d1 && a < d2 ? vmask2 : 0);
						if (fill) {
							for (uint8_t i = 1; i <= n; ++i) {
								*(a + i) = 0xff;
							}
						}
						break;
					case XOR:
						*a ^= mask1 | (fill && a > d1 && a < d2 ? vmask1 : 0);
						*b ^= mask2 | (fill && a > d1 && a < d2 ? vmask2 : 0);
						if (fill) {
							for (uint8_t i = 1; i <= n; ++i) {
								*(a + i) ^= 0xff;
							}
						}
						break;
					case CLEAR:
						*a &= ~(mask1 | (fill && a > d1 && a < d2 ? vmask1 : 0));
						*b &= ~(mask2 | (fill && a > d1 && a < d2 ? vmask2 : 0));
						if (fill) {
							for (uint8_t i = 1; i <= n; ++i) {
								*(a + i) = 0x00;
							}
						}
						break;
					default:
						;
				}
			}
			a -= h;
			b -= h;
			if (fill) n = 0;

			//Border
			switch (mode) {
				case COPY:
				case OR:
					*d1 |= vmask1;
					*d3 |= vmask2;
					*a |= vmask1;
					*b |= vmask2;
					for (uint8_t i = 1; i <= n; ++i) {
						*(d1 + i) = 0xff;
						*(a + i) = 0xff;
					}
					break;
				case XOR:
					*d1 ^= vmask1;
					*d3 ^= vmask2;
					*a ^= vmask1;
					*b ^= vmask2;
					for (uint8_t i = 1; i <= n; ++i) {
						*(d1 + i) ^= 0xff;
						*(a + i) ^= 0xff;
					}
					break;
				case CLEAR:
					*d1 &= ~vmask1;
					*d3 &= ~vmask2;
					*a &= ~vmask1;
					*b &= ~vmask2;
					for (uint8_t i = 1; i <= n; ++i) {
						*(d1 + i) = 0x00;
						*(a + i) = 0x00;
					}
					break;
				default:
					;
			}
		#endif
	#endif						
}

void Disp_Line(const uint8_t x1,
			   const uint8_t y1,
			   const uint8_t x2,
			   const uint8_t y2,
			   const draw_mode_t mode) {

	if (x1 > DISP_WIDTH - 1 || x2 > DISP_WIDTH - 1) return;
	if (y1 > DISP_HEIGHT - 1 || y2 > DISP_HEIGHT - 1) return;
	if (x1 == x2 && y1 == y2) {
		switch (mode) {
			case COPY:
			case OR:
				Disp_SetPx(x1, y1); return;
			case XOR:
				Disp_InvPx(x1, y1); return;
			case CLEAR:
				Disp_ClrPx(x1, y1); return;
			default: return;
		}
	}

	uint8_t ix1, iy1, ix2, iy2;
	//Direction is always from left to right
	if (x1 <= x2) {
		ix1 = x1;
		iy1 = y1;
		ix2 = x2;
		iy2 = y2;
	} else {
		ix1 = x2;
		iy1 = y2;
		ix2 = x1;
		iy2 = y1;
	}

	#if DISP_USE_CACHE == 1 && DISP_CACHE_ALGN == 1
		if (iy1 == iy2) {
			//Horizontal line
			uint8_t *d1, *d2, mask;
			d1 = CacheXY(ix1, iy1, &mask);
			d2 = CacheXY(ix2, iy2, &mask);
			for (; d1 <= d2; d1 += DISP_HEIGHT / 8) {
				UseMask(d1, mask, mode);
			}
		} else if (ix1 == ix2) {
			//Vertical line
			if (iy2 < iy1) {
				iy1 = y2;
				iy2 = y1;
			}
			uint8_t *d1, *d2, mask1, mask2;
			d1 = CacheXY(ix1, iy1, &mask1);
			d2 = CacheXY(ix2, iy2, &mask2);
			if (d1 == d2) {
				mask1 = ~((mask1 - 1) | ~(mask2 | (mask2 - 1)));
				UseMask(d1, mask1, mode);
			} else {
				uint8_t n = d2 - d1 - 1;
				if (n == 0xff) n = 0;
				mask1 = ~(mask1 - 1);
				mask2 = (mask2 - 1) | mask2;
				switch (mode) {
					case COPY:
					case OR:
						*d1++ |= mask1;
						*d2 |= mask2;
						for (; n; --n, ++d1) {
							*d1 |= 0xff;
						}
						break;
					case XOR:
						*d1++ ^= mask1;
						*d2 ^= mask2;
						for (; n; --n, ++d1) {
							*d1 ^= 0xff;
						}
						break;
					case CLEAR:
						*d1++ &= ~mask1;
						*d2 &= mask2;
						for (; n; --n, ++d1) {
							*d1 = 0;
						}
						break;
					default:
						;
				}
			}
		} else {
			//No orthogonal line - Greets from Bresenham
			uint8_t dx = ix2 - ix1, dy;
			bool dir;
			if (iy1 < iy2) {
				dy = iy2 - iy1;
				dir = false;
			} else {
				dy = iy1 - iy2;
				dir = true;
			}

			uint8_t *d1, *d2, mask;
			d2 = CacheXY(ix2, iy2, &mask);
			d1 = CacheXY(ix1, iy1, &mask);

			int8_t err;
			if (dx > dy) {
				//Angle is less than 45
				UseMask(d1, mask, mode);
				err = dx / 2;
				for (uint8_t i = 0; i < dx; ++i) {
					err -= dy;
					if (err < 0) {
						if (dir) {
							mask >>= 1;
							if (!mask) {
								--d1;
								mask = 0x80;
							}
						} else {
							mask <<= 1;
							if (!mask) {
								++d1;
								mask = 0x01;
							}
						}
						err += dx;
					}
					d1 += DISP_HEIGHT / 8;
					UseMask(d1, mask, mode);
				}
			} else {
				//Angle is greater than or equal to 45
				err = dy / 2;
				uint8_t b = mask;

				for (uint8_t i = 0; i < dy; ++i) {
					err -= dx;
					if (err < 0) {
						UseMask(d1, mask, mode);
						d1 += DISP_HEIGHT / 8;
						err += dy;
						mask = 0x00;
					}
					if (dir) {
						b >>= 1;
						if (!b) {
							UseMask(d1, mask, mode);
							--d1;
							b = 0x80;
							mask = 0x80;
						} else {
							mask |= b;
						}
					} else {
						b <<= 1;
						if (!b) {
							UseMask(d1, mask, mode);
							++d1;
							b = 0x01;
							mask = 0x01;
						} else {
							mask |= b;
						}
					}
				}
				UseMask(d1, mask, mode);
			}
		}
	#endif
}
#endif

#if DISP_USE_CACHE == 1
void Disp_Clear_Cache() {
	for (uint16_t i = 0; i < DISP_WIDTH * DISP_HEIGHT / 8; ++i) {
		Cache[i] = 0;
	}
}

void Disp_Refresh() {
	DISP_SELECT;
	DISP_DC_CMD;
	SendSPI(0x22);	//Startup, Vertical addressing, Basic instruction set
	SendSPI(0x40);	//Y = 0;
	SendSPI(0x80);	//X = 0;
	DISP_DC_DATA;
	for (uint16_t i = 0; i < DISP_WIDTH * DISP_HEIGHT / 8; ++i) {
		loop_until_bit_is_set(SPSR, SPIF);
		SPDR = Cache[i];
	}
	loop_until_bit_is_set(SPSR, SPIF);
	DISP_RELEASE;
}
#endif

void Disp_Contrast(const uint8_t contrast) {
	SendSPI(0x23);
	SendSPI(0x80 | contrast);
	SendSPI(0x22);
}





