/******************************************************************************
* Project:		morsix3330
* File:			printf.c
* Date:			25.12.2007
* Author:		Philipp Burch
* E-Mail:		philipp_burch@gmx.ch
* Description:	This file contains a special version of the printf standard
*				function to reduce the overhead introduced by the normal stdio
*				functions.
******************************************************************************/

#include <printf.h>
#include <stdlib.h>
#include <disp3330.h>
#include <avr/io.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <avr/pgmspace.h>
#include <debug.h>


printf_target_t Target = USART;
uint8_t Disp_Left, Disp_Top, Disp_Width = DISP_WIDTH, Disp_Height = DISP_HEIGHT;
point_t Disp_XY;
draw_mode_t Disp_Draw_Mode = COPY;
bool Disp_Break_Word = false;
bool Disp_Use_Controls = true;
bool Disp_Auto_Refresh = true;

void PrintF_Target(const printf_target_t tg) {
	Target = tg;
}

void PrintF_Disp_Margins(const uint8_t left,
						 const uint8_t top,
						 const uint8_t width,
						 const uint8_t height) {

	Disp_Left = left;
	Disp_Top = top;
	Disp_Width = width;
	Disp_Height = height;
}

void PrintF_Disp_Locate(const uint8_t x, const uint8_t y) {
	Disp_XY.x = x;
	Disp_XY.y = y;
}

void PrintF_Disp_Ctrl(const draw_mode_t draw_mode,
					  const bool break_word,
					  const bool use_controls,
					  const bool auto_refresh) {

	Disp_Draw_Mode = draw_mode;
	Disp_Break_Word = break_word;
	Disp_Use_Controls = use_controls;
	Disp_Auto_Refresh = auto_refresh;
}

void PutS(char *s) {
	if (Target & USART) {
		uint8_t i = 0;
		while (s[i]) {
			loop_until_bit_is_set(UCSR0A, UDRE0);
			UDR0 = s[i++];
		}
	}
	if (Target & DISPLAY) {
		if (Disp_Auto_Refresh) Disp_Clear_Cache();
		Disp_XY = Disp_Print(Disp_Left, 
							 Disp_Top, 
							 Disp_Width, 
							 Disp_Height, 
							 Disp_XY.x, 
							 Disp_XY.y, 
							 s, 
							 Disp_Draw_Mode, 
							 Disp_Break_Word, 
							 Disp_Use_Controls);

		if (Disp_Auto_Refresh) Disp_Refresh();
	}
	if (Target & SPEAKER) {
		
	}
	if (Target & PHONE) {
		
	}
}

int16_t PrintF(const char *format, ...) {
	if (*format == 0) return -1;

	uint8_t bcnt = 0;
	uint16_t fi = 0;
	uint16_t written = 0;
	static char buffer[PRINTF_BUF_SIZE + 1];
	buffer[PRINTF_BUF_SIZE] = 0;
	char *buf = buffer;

	printf_flag_t flags;
	uint8_t width;
	uint8_t precision;
	uint8_t length;
	char specifier;
	uint8_t cnt;
	bool vastart = false;
	va_list vl;

	while (format[fi]) {
		if (format[fi] != '%') {
			*buf++ = format[fi++];
			if (++bcnt == PRINTF_BUF_SIZE - 2) {
				buffer[bcnt] = 0;
				PutS(buffer);
				written += bcnt;
				buf = buffer;
				bcnt = 0;
			}
		} else {
			++fi;
			if (format[fi] == '%') {
				*buf = '%';
				++bcnt;
				++fi;
				continue;
			}

			//Format specifier

			if (!vastart) {
				va_start(vl, format);
				vastart = true;
			}

			flags = 0;
			width = 0;
			precision = 1;
			length = 2;

			//Flags
			do {
				cnt = 0;
				if (format[fi] == '-') {
					flags |= LEFT_JUSTIFY;
					cnt = 1;
				}
				if (format[fi] == '+') {
					flags |= FORCE_SIGN;
					cnt = 1;
				}
				if (format[fi] == ' ') {
					flags |= USE_PADDING_SPACE;
					cnt = 1;
				}
				if (format[fi] == '#') {
					flags |= USE_PREFIX;
					cnt = 1;
				}
				if (format[fi] == '0') {
					flags |= USE_PADDING_ZERO;
					cnt = 1;
				}
				if (cnt) ++fi;
			} while (cnt && format[fi]);

			//Width
			if (format[fi] == '*') {
				width = va_arg(vl, uint16_t);
				++fi;
			} else if (isdigit(format[fi])) {
				width = format[fi++] - '0';
				if (isdigit(format[fi])) {
					width += (format[fi++] - '0') * 10;
				}
			}
			if (width > 16) width = 16;

			//Precision
			if (format[fi] == '.') {
				++fi;
				if (format[fi] == '*') {
					precision = va_arg(vl, uint16_t);
					++fi;
				} else if (isdigit(format[fi])) {
					precision = format[fi++] - '0';
					if (isdigit(format[fi])) {
						precision += (format[fi++] - '0') * 10;
					}
				} else {
					precision = 0;
				}
			}
			if (precision > 16) precision = 16;

			//Length
			if (format[fi] == 'l') {
				length = 4;
				++fi;
			} else if (format[fi] == 'h') {
				//Standard, anyway...
				length = 2;
				++fi;
			}
			
			//Format specifier (Must be present)
			specifier = format[fi++];

			//Make room for the result (At least 20 chars)
			if (PRINTF_BUF_SIZE - bcnt < 20) {
				buffer[bcnt] = 0;
				PutS(buffer);
				written += bcnt;
				buf = buffer;
				bcnt = 0;
			}

			int32_t val;
			int8_t n;
			uint8_t len;
			char *start;
				
			switch (specifier) {
				case 'c':
					//Character
					*buf++ = (char)va_arg(vl, uint16_t);
					++bcnt;
					break;
				case 'd':
				case 'i':
				case 'o':
				case 'u':
				case 'x':
				case 'X':
					//Integer
					if (length == 2) {
						val = va_arg(vl, int16_t);
					} else {
						val = va_arg(vl, int32_t);
					}

debug_send_var("val", val);

					start = buf;

					if (specifier == 'd' || specifier == 'i' || specifier == 'u') {
						if (flags & FORCE_SIGN || flags & USE_PADDING_SPACE) {
							if (val >= 0 && specifier != 'u') {
								if (flags & USE_PADDING_SPACE) {
									*buf++ = ' ';
								} else {
									*buf++ = '+';
								}
							}
						}
						if (specifier == 'o' && flags & USE_PREFIX) *buf++ = '0';
						if (specifier == 'u' && length == 2) {
							itoa((uint16_t)val, buf, specifier == 'o' ? 8 : 10);
						} else {
							if (specifier == 'u') {
								ltoa((uint32_t)val, buf, specifier == 'o' ? 8 : 10);
							} else {
								ltoa(val, buf, specifier == 'o' ? 8 : 10);
							}
						}
					} else if (specifier == 'x' || specifier == 'X' || specifier == 'o') {
						if (flags & USE_PREFIX) {
							*buf++ = '0';
							if (specifier != 'o') *buf++ = specifier;
						}

						if (precision || val) {
							if (length == 2) {
								itoa((uint16_t)val, buf, specifier == 'o' ? 8 : 16);
							} else {
								ltoa(val, buf, specifier == 'o' ? 8 : 16);
							}
							if (specifier == 'X') strupr(buf - (flags & USE_PREFIX ? 1 : 0));
						} else {
							*buf = 0;
						}
					}

					bcnt += buf - start;
					
					n = precision - strlen(buf);
					if (n < 0) n = 0;
					len = strlen(buf);

					if (n) {
						for (int8_t i = len + n; i >= n; --i) {
							buf[i] = buf[i - n];
						}
						for (uint8_t i = 0; i < n; ++i) {
							buf[i] = '0';
						}
					}

					buf += len;
					bcnt += len;

					len += buf - start + n;
					n = width - len;
					if (n < 0) n = 0;

					if (n) {
						if (flags & LEFT_JUSTIFY) {
							for (uint8_t i = len; i < len + n; ++i) {
								start[i] = ' ';
							}
						} else {
							for (int8_t i = len + n; i >= n; --i) {
								start[i] = start[i - n];
							}
							for (uint8_t i = 0; i < n; ++i) {
								start[i] = flags & USE_PADDING_ZERO ? '0' : ' ';
							}
						}
						len += n;
					}
					break;

				case 'p':
					//Pointer (16 Bit integer, format hh:ll
					buf[0] = '0'; buf[1] = '0'; buf[2] = '0';
					uint16_t val = (uint16_t)va_arg(vl, void *);
					n = 0;
					if (val < 0x1000) ++n;
					if (val < 0x0100) ++n;
					if (val < 0x0010) ++n;
					itoa(val, buf + n, 16);
					buf[4] = buf[3];
					buf[3] = buf[2];
					buf[2] = ':';
					buf += 5;
					bcnt += 5;
					break;

				case 'f':
					//Fixed point number
					if (length == 2) {
						int16_t v;
						v = va_arg(vl, int16_t);
						val = v;
					} else {
						val = va_arg(vl, int32_t);
					}

debug_send_var("val", val);

					start = buf;

					if (val >= 0 && flags & FORCE_SIGN) {
						if (flags & USE_PADDING_SPACE) {
							*buf++ = ' ';
						} else {
							*buf++ = '+';
						}
					}

					bcnt += buf - start;
				
					if (precision || val) {
						if (length == 2) {
							itoa((int16_t)val, buf, 10);
						} else {
							ltoa(val, buf, 10);
						}
					} else {
						*buf = 0;
					}

debug_send_text_var("buf", buf, 0);

					if (val < 0) ++buf;

					len = strlen(buf);

					int8_t dp = len - precision;
					if (dp < 0) dp = 0;
					
					n = precision - len;
					if (n < 0) n = 0;
					++n;
					for (int8_t i = len; i >= dp; --i) {
						buf[i + n] = buf[i];
					}
					if (precision > len) {
						for (uint8_t i = 0; i < n - 1; ++i) {
							buf[i + 1] = '0';
						}
					}
					len += n;
					buf[dp] = '.';

					buf += len;
					bcnt += len;

					len += buf - start;

					n = width - len;
					if (n < 0) n = 0;

					if (n) {
						if (flags & LEFT_JUSTIFY) {
							for (uint8_t i = len; i < len + n; ++i) {
								start[i] = ' ';
							}
						} else {
							for (int8_t i = len + n; i >= n; --i) {
								start[i] = start[i - n];
							}
							for (uint8_t i = 0; i < n; ++i) {
								start[i] = flags & USE_PADDING_ZERO ? '0' : ' ';
							}
						}
						len += n;
					}
					break;

				case 's':
					//String
					if (bcnt) {
						buffer[bcnt] = 0;
						PutS(buffer);
						written += bcnt;
						buf = buffer;
						bcnt = 0;
					}

					char *s = va_arg(vl, char *);

					written += strlen(s);

					PutS(s);
					break;

				case 'S':
					//String located in Flash-ROM
					if (bcnt) {
						buffer[bcnt] = 0;
						PutS(buffer);
						written += bcnt;
						buf = buffer;
						bcnt = 0;
					}

					prog_char *sp = va_arg(vl, prog_char *);

					do {
						strncpy_P(buffer, sp, PRINTF_BUF_SIZE);
						written += strlen(buffer);
						PutS(buffer);
					} while (buffer[PRINTF_BUF_SIZE - 1]);
					break;
			}
		}
	}
	buffer[bcnt] = 0;
	if (bcnt) PutS(buffer);
	written += bcnt;

	if (vastart) va_end(vl);

	return written;
}

